/*------------------------------------------------------------------------*/
/* Copyright (C) SIEMENS CORP., 2018 All rights reserved.*/
/* All rights reserved.                                                   */
/* Redistribution and use in source and binary forms, with or without     */
/* modification, are permitted provided that the following conditions     */
/* are met:                                                               */
/* 1. Redistributions of source code must retain the above copyright      */
/* notice, this list of conditions and the following disclaimer.          */
/* 2. Redistributions in binary form must reproduce the above copyright   */
/* notice, this list of conditions and the following disclaimer in the    */
/* documentation and/or other materials provided with the distribution.   */
/* 3. Neither the name of the copyright holder nor the names of its       */
/* contributors may be used to endorse or promote products derived from   */
/* this software without specific prior written permission.               */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      */
/* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS      */
/* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE         */
/* COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   */
/* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,   */
/* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS  */
/* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)                    */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,    */
/* STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING  */
/* IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE     */
/* POSSIBILITY OF SUCH DAMAGE.                                            */
/*------------------------------------------------------------------------*/
/*------------------------------------------------------------------------*/
/* Attention : Callbacks are running concurrent in other threads so all   */
/*             printf statements should be synchronized. But this sample  */
/*             application doesn't synchronize for simplicity  !          */
/*------------------------------------------------------------------------*/

#include "os/criticalsection.h"

#include <stdlib.h>

#if defined(_WIN32)
#include <Windows.h>
#elif defined(__unix)
#include <pthread.h>
#endif

typedef struct OsCriticalSection
{
#if defined(_WIN32)
	CRITICAL_SECTION win_cs;
#elif defined(__unix)
	pthread_mutex_t lin_cs;
#endif
}OsCriticalSection;

/* Initialize the Critical Section */
void CriticalSectionCreate(CriticalSection* cs)
{
	cs->os_cs = (OsCriticalSection*)malloc(sizeof(OsCriticalSection));
#if defined(_WIN32)
	InitializeCriticalSection(&cs->os_cs->win_cs);
#elif defined(__unix)
	pthread_mutex_init(&cs->os_cs->lin_cs, 0);
#endif
}

/* Deinitialize the Critical Section */
void CriticalSectionDelete(CriticalSection* cs)
{
#if defined(_WIN32)
	DeleteCriticalSection(&cs->os_cs->win_cs);
#elif defined(__unix)
	pthread_mutex_destroy(&cs->os_cs->lin_cs);
#endif
	free(cs->os_cs);
}

/* Enter the Critical Section */
void CriticalSectionEnter(CriticalSection* cs)
{
#if defined(_WIN32)
	EnterCriticalSection(&cs->os_cs->win_cs);
#elif defined(__unix)
	pthread_mutex_lock(&cs->os_cs->lin_cs);
#endif
}

/* Leave the Critical Section */
void CriticalSectionLeave(CriticalSection* cs)
{
#if defined(_WIN32)
	LeaveCriticalSection(&cs->os_cs->win_cs);
#elif defined(__unix)
	pthread_mutex_unlock(&cs->os_cs->lin_cs);
#endif
}
